# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

__version__ = '0.11.0'
version_info = (0, 11, 0)

__title__ = 'fitdecode'
__fancy_title__ = 'fitdecode'
__description__ = 'FIT file parser and decoder'
__url__ = 'https://github.com/polyvertex/fitdecode'
__license__ = 'MIT'
__author__ = 'Jean-Charles Lefebvre'
__author_email__ = 'polyvertex@gmail.com'
__copyright__ = 'Copyright 2018-present Jean-Charles Lefebvre'

__keywords__ = ['fit', 'ant', 'file', 'parse', 'parser', 'decode', 'decoder']
